class FileAware:

    def files_list(self): ...

    def file_get(self, **kwargs): ...

    def upload(self, **kwargs): ...

    def download(self, **kwargs): ...

    def files_delete(self, **kwargs): ...

    def nuke(self): ...



class ModelAware:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model: str | None = None


class Connectable:
    def connect(self) -> bool: ...

    def close(self): ...

    def __enter__(self):
        assert self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from typing import Literal

from openai import OpenAI, AuthenticationError, PermissionDeniedError
from openai.types import Model
from openai.types.chat import ChatCompletion

from davidkhala.llm.model import Connectable
from davidkhala.llm.model.chat import on_response, ChoicesChat
from davidkhala.llm.model.embed import EmbeddingAware
from davidkhala.llm.model.garden import GardenAlike


class Client(ChoicesChat, EmbeddingAware, GardenAlike, Connectable):

    def __init__(self, client: OpenAI):
        super().__init__()
        self.client: OpenAI = client
        self.encoding_format: Literal["float", "base64"] = "float"

    def connect(self):
        try:
            self.list_models()
            return True
        except AuthenticationError | PermissionDeniedError:
            return False

    def list_models(self) -> list[Model]:
        return self.client.models.list().data

    def encode(self, *_input: str) -> list[list[float]]:
        response = self.client.embeddings.create(
            model=self.model,
            input=list(_input),
            encoding_format=self.encoding_format
        )
        return [item.embedding for item in response.data]

    def chat_create(self, *user_prompt, **kwargs) -> ChatCompletion:
        return self.client.chat.completions.create(
            model=self.model,
            messages=self.messages_from(*user_prompt),
            n=self.n,
            **kwargs
        )

    def chat(self, *user_prompt, **kwargs):
        response: ChatCompletion = self.chat_create(*user_prompt, **kwargs)
        return on_response(response, self.n)

    def close(self):
        self.client.close()



from openai import OpenAI
from davidkhala.llm.openai import Client as BaseClient


class Client(BaseClient):
    def __init__(self, o: str | int, token: str):
        """
        :param o: databricks organization id. Appears as parameter of workspace url
        """
        super().__init__(OpenAI(
            base_url=f"https://{o}.ai-gateway.cloud.databricks.com/mlflow/v1",
            api_key=token
        ))

    def chat(self, *user_prompt, **kwargs):
        """Databricks always reasoning, and content is not of type str"""
        contents:list[list[dict]] = super().chat(*user_prompt, **kwargs)

        for content in contents:
            for block in content:
                match block['type']:
                    case "text":
                        yield block['text']
                    case "reasoning":
                        for s in block["summary"]:
                            assert s['type'] == 'summary_text'
                            yield s['text']


from pathlib import Path
from typing import Literal

from openai import OpenAI
from openai.types import FileObject

from davidkhala.llm.model.chat import MessageDict
from davidkhala.llm.model.file import FileAware
from davidkhala.llm.openai import Client as BaseClient
from davidkhala.utils.syntax.format import mime_of


class Client(BaseClient, FileAware):

    def __init__(self, api_key: str, base_url: str):
        super().__init__(OpenAI(api_key=api_key, base_url=base_url))

    def upload(self, file: Path, purpose: Literal["file-extract", "image", "video"] | None = None) -> FileObject:
        """
        Supported file formats:
        .pdf, .txt, .csv, .doc, .docx, .xls, .xlsx, .ppt, .pptx, .md, .jpeg, .png, .bmp, .gif, .svg, .svgz, .webp, .ico, .xbm, .dib, .pjp, .tif, .pjpeg, .avif, .dot, .apng, .epub, .tiff, .jfif, .html, .json, .mobi, .log, .go, .h, .c, .cpp, .cxx, .cc, .cs, .java, .js, .css, .jsp, .php, .py, .py3, .asp, .yaml, .yml, .ini, .conf, .ts, .tsx,
        """
        if not purpose:
            match mime_of(file):
                case "application/pdf":
                    purpose = 'file-extract'
        return self.client.files.create(file=file, purpose=purpose)

    def file_get(self, file_id: str) -> FileObject:
        return self.client.files.retrieve(file_id)

    def files_list(self):
        return self.client.files.list().data

    def download(self, file_id: str) -> str:

        return self.client.files.content(file_id).text

    def files_delete(self, file_id: str):
        self.client.files.delete(file_id=file_id)

    def message_of(self, file: Path) -> MessageDict:
        file_id = self.upload(file).id
        return {
            'role': 'system',
            'content': self.download(file_id)
        }

    def nuke(self):
        files = self.files_list()
        for file in files:
            self.files_delete(file.id)


class Global(Client):
    def __init__(self, api_key: str):
        super().__init__(api_key, "https://api.moonshot.ai/v1")



from abc import abstractmethod, ABC
from pathlib import Path
from typing import Protocol, Any, Iterable, TypedDict, Literal, NotRequired

from davidkhala.utils.syntax.format import Base64, mime_of
from davidkhala.utils.syntax.url import filename_from
from pydantic import BaseModel

from davidkhala.llm.model import ModelAware


class MessageProtocol(Protocol):
    content: str | list | Any | None
    role: str


class ChoiceProtocol(Protocol):
    message: MessageProtocol


class ChoicesAware(Protocol):
    choices: list[ChoiceProtocol]


class MultimodalPrompt(BaseModel):
    text: str


class ImagePrompt(MultimodalPrompt):
    image_url: list[str]
    detail: Literal['auto', 'low', 'high'] | None = 'auto'  # Image detail level for vision models


class FilePrompt(MultimodalPrompt):
    url: list[str] | None = None
    path: list[Path] | None = None


Prompt = str | ImagePrompt | FilePrompt


def on_response(response: ChoicesAware, n: int | None):
    contents = [choice.message.content for choice in response.choices]
    if n:
        assert len(contents) == n, f"expected {n} choices, but got {len(contents)}"
    return contents


class ImageContentPart(TypedDict):
    type: Literal['image_url']
    image_url: NotRequired[dict]


class TextContentPart(TypedDict):
    type: Literal['text']
    text: str


ContentPart = TextContentPart | ImageContentPart


class FileAnnotation(TypedDict):
    hash: str
    name: str | None
    content: list[ContentPart]


class AnnotationDict(TypedDict):
    type: Literal['file']
    file: FileAnnotation


class MessageDict(TypedDict):
    content: NotRequired[str | list]
    role: str
    annotations: NotRequired[list[AnnotationDict]]


def messages_from(*user_prompt: Prompt) -> Iterable[MessageDict]:
    for _ in user_prompt:
        message = MessageDict(role='user')
        match _:
            case str():
                message['content'] = _
            case MultimodalPrompt():
                message['content'] = [{"type": "text", "text": _.text}]
                match _:
                    case ImagePrompt():
                        message['content'].extend(
                            {"type": "image_url", "image_url": {"url": i, 'detail': _.detail}} for i in _.image_url)

                    case FilePrompt():
                        if _.url:
                            message['content'].extend({"type": "file", "file": {
                                "filename": filename_from(item), "file_data": item
                            }} for item in _.url)

                        if _.path:
                            for item in _.path:
                                message['content'].append({"type": "file", "file": {
                                    "filename": item.name,
                                    "file_data": f"data:{mime_of(item)};base64,{Base64.encode_file(item)}"
                                }})

        yield message


class ChatAware(ABC, ModelAware):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.messages: list[Any | MessageDict] = []

    def reset(self):
        self.messages = []

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.messages = [MessageDict(role='system', content=sys_prompt)]

    @abstractmethod
    def chat(self, *user_prompt, **kwargs): ...

    def messages_from(self, *user_prompt: Prompt):
        self.messages.extend(messages_from(*user_prompt))
        return self.messages

    def for_next(self, message: MessageProtocol | MessageDict):
        self.messages.append(message)

    def with_annotations(self, annotations: list[AnnotationDict]):
        # file should not be excluded from message thread. Just openrouter will skip parsing costs
        self.for_next(MessageDict(role="assistant", annotations=annotations))


class ChoicesChat(ChatAware, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.n: int = 1

    def reset(self):
        super().reset()
        self.n = 1


class CompareChatAware(ChatAware, ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._models: list[str] = []

    def as_chat(self, *models: str, sys_prompt: str = None):
        if len(models) > 1:
            self._models = list(models)
            super().as_chat(None, sys_prompt)
        elif len(models) == 1:
            super().as_chat(models[0], sys_prompt)



from abc import ABC
from typing import Literal

from davidkhala.llm.model.chat import CompareChatAware
from davidkhala.llm.model.garden import GardenAlike


class OpenRouterModel(CompareChatAware,GardenAlike, ABC):
    """
    openrouter has no `n` parameter support
    """
    n = 1  # only has one fake choice. Openrouter use models as pool for load-balance only


class Plugins:
    PDF_ENGINE = Literal['pdf-text', 'mistral-ocr', 'native'] | None

    @staticmethod
    def pdf(engine: PDF_ENGINE = 'pdf-text'):
        """
        PDF parsing will still work even if the plugin is not explicitly set
        :param engine:
            'pdf-text': free. Best for well-structured PDFs with clear text content
            'native': Only available for models that support file input natively. Token cost
            'mistral-ocr': default. Best quality. $2 per 1,000 pages
        """
        return {'id': 'file-parser', 'pdf': {'engine': engine}}



import datetime
from abc import ABC

from davidkhala.utils.http_request import Request
from davidkhala.utils.typed import IDDict

from davidkhala.llm.model.chat import ChatAware, Prompt
from davidkhala.llm.model.embed import EmbeddingAware
from davidkhala.llm.model.garden import GardenAlike


class API(Request):
    def __init__(self, api_key: str, base_url: str):
        super().__init__({
            "bearer": api_key
        })
        self.base_url = base_url + '/v1'


class ChatAPI(API, ChatAware):

    def chat(self, *user_prompt: Prompt, **kwargs):

        json = {
            "messages": self.messages_from(*user_prompt),
            **kwargs,
        }

        response = self.request(f"{self.base_url}/chat/completions", "POST", json=json)

        data = []
        file_annotations = []
        for _ in response['choices']:
            data.append(_['message']['content'])
            _a = _["message"].get('annotations')
            if _a: file_annotations.append(_a)

        return {
            "data": data,
            "meta": {
                "usage": response['usage'],
                "created": datetime.datetime.fromtimestamp(response['created']),
            },
            'model': response['model'],
            'annotations': file_annotations,
        }


class EmbeddingAPI(API, EmbeddingAware):
    def encode(self, *_input: str) -> list[list[float]]:
        response = self.request(f"{self.base_url}/embeddings", "POST", json={
            'input': _input,
            'model': self.model
        })
        return [_['embedding'] for _ in response['data']]


class GardenAPI(API, GardenAlike, ABC):
    def list_models(self) -> list[IDDict]:
        response = self.request(f"{self.base_url}/models", "GET")
        return response['data']



from abc import ABC, abstractmethod

from davidkhala.utils.protocol import ID


class GardenAlike(ABC):
    @abstractmethod
    def list_models(self) -> list[ID]: ...

    @property
    def models(self) -> list[str]:
        return [m.id for m in self.list_models()]


class TrialAvailable(ABC):
    @property
    @abstractmethod
    def free_models(self) -> list[str]: ...



from openrouter.components import Model
from openrouter.errors import UnauthorizedResponseError
from openrouter.operations import ListData, CreateEmbeddingsResponseBody
from typing_extensions import Literal

from davidkhala.llm.model import Connectable
from davidkhala.llm.model.chat import on_response, Prompt
from davidkhala.llm.model.embed import EmbeddingAware
from davidkhala.llm.model.garden import TrialAvailable
from davidkhala.llm.model.openrouter import OpenRouterModel
from openrouter import OpenRouter


class Client(OpenRouterModel, EmbeddingAware, Connectable, TrialAvailable):
    @property
    def free_models(self) -> list[str]:
        return ['openrouter/free', *(_.id for _ in self.list_models() if _.id.endswith(':free'))]

    def __init__(self, api_key: str):
        super().__init__()
        self.client: OpenRouter = OpenRouter(api_key)

    def chat(self, *user_prompt: Prompt) -> str:
        """
        python SDK Do not support FilePrompt yet
        """
        plugins = []
        r = self.client.chat.send(
            model=self.model,
            models=self._models,
            messages=self.messages_from(*user_prompt),
            plugins=plugins
        )
        return on_response(r, OpenRouterModel.n)[0]

    def connect(self):
        try:
            _ = self.models
            return True
        except UnauthorizedResponseError:
            return False

    def list_models(self, _type: Literal['embeddings'] | None = None) -> list[Model]:
        match _type:
            case 'embeddings':
                return self.client.embeddings.list_models().data
        return self.client.models.list().data

    def encode(self, *_input: str) -> list[list[float]]:
        r: CreateEmbeddingsResponseBody = self.client.embeddings.generate(
            input=_input,
            model=self.model
        )
        return [_.embedding for _ in r.data]


class Admin:
    def __init__(self, provisioning_key: str):
        super().__init__()
        self.provisioning_key = provisioning_key
        self.client = OpenRouter(provisioning_key)

    @property
    def keys(self) -> list[ListData]:
        return self.client.api_keys.list().data



from anthropic import Anthropic, __version__
from anthropic.types import Message

from davidkhala.llm.model.chat import ChatAware, Prompt, MessageDict
from davidkhala.llm.model.file import FileAware

version = __version__


class Client(ChatAware, FileAware):
    def __init__(self, *,
                 api_key: str | None = None,
                 base_url: str | None = None, ):
        super().__init__()
        self.client = Anthropic(api_key=api_key, base_url=base_url)
        self.system: str | None = None  # system prompt
        self.max_tokens = 1024

    def chat(self, *user_prompt: Prompt):
        r: Message = self.client.messages.create(
            model=self.model,
            system=self.system,
            messages=self.messages_from(*user_prompt),
            max_tokens=self.max_tokens,  # required
        )
        self.messages.append(MessageDict(content=r.content, role=r.role))
        return [{'text': _.text, 'citations': _.citations} for _ in r.content if _.type == 'text']

    def as_chat(self, model: str | None, sys_prompt: str = None):
        self.model = model
        if sys_prompt is not None:
            self.system = sys_prompt



from davidkhala.llm.model.chat import MessageDict
from davidkhala.llm.openai import Client as BaseClient, OpenAI, on_response


class Client(BaseClient):
    n = 1  # minimax does not support n

    def chat(self, *user_prompt, **kwargs) -> tuple[str, list[dict]]:
        kwargs['extra_body'] = {'reasoning_split': True}
        response = super().chat_create(*user_prompt, **kwargs)

        assert len(response.choices) == Client.n, "minimax does not support n"
        message = response.choices[0].message

        self.for_next(MessageDict(content=message.content, role=message.role))

        return message.content, message.reasoning_details  # reasoning_details is hidden field of minimax


class Global(Client):
    def __init__(self, api_key: str):
        super().__init__(OpenAI(api_key=api_key, base_url='https://api.minimax.io/v1'))


class CN(Client):
    def __init__(self, api_key: str):
        super().__init__(OpenAI(api_key=api_key, base_url='https://api.minimaxi.com/v1'))



from davidkhala.utils.http_request import default_on_response, Request
from requests import Response, HTTPError as RawHTTPError

from davidkhala.llm.api import GardenAPI, ChatAPI, EmbeddingAPI
from davidkhala.llm.model.chat import ImagePrompt, ChoicesChat
from davidkhala.llm.model.garden import TrialAvailable
from davidkhala.llm.model.rerank import Reranker

Prompt = str | ImagePrompt


class HTTPError(RawHTTPError):
    @staticmethod
    def from_raw(err: RawHTTPError):
        return HTTPError(err.args[0], response=err.response)

    @property
    def code(self):
        return self.json['code']

    @property
    def json(self):
        return self.response.json()

    @property
    def message(self):
        return self.json['message']

class Console(Request):
    # TODO siliconflow cn use authenticator login
    def price_of(self, model_id):
        # TODO login is required
        url = f"https://cloud.siliconflow.cn/biz-server/api/v1/playground/{model_id}/biz_info"
        r = self.request(url, 'GET')
        return r

class SiliconFlow(ChatAPI, ChoicesChat, Reranker, EmbeddingAPI, GardenAPI, TrialAvailable):
    @property
    def free_models(self) -> list[str]:
        """
        Cannot be lively fetched by list_models
        """
        return [
            # chat section
            'THUDM/GLM-4.1V-9B-Thinking'
            'THUDM/GLM-Z1-9B-0414'
            'THUDM/GLM-4-9B-0414'
            'THUDM/glm-4-9b-chat'
            'Qwen/Qwen3-8B'
            'Qwen/Qwen2.5-7B-Instruct'
            'Qwen/Qwen2.5-Coder-7B-Instruct'
            'internlm/internlm2_5-7b-chat'
            'deepseek-ai/DeepSeek-R1-0528-Qwen3-8B',
            'deepseek-ai/DeepSeek-R1-Distill-Qwen-7B',
            # embedding and reranker
            'BAAI/bge-m3'
            'BAAI/bge-reranker-v2-m3'
            'BAAI/bge-large-zh-v1.5'
            'BAAI/bge-large-en-v1.5'
            'netease-youdao/bce-reranker-base_v1'
            'netease-youdao/bce-embedding-base_v1'
            # Audio
            'FunAudioLLM/SenseVoiceSmall'
            # image
            'Kwai-Kolors/Kolors'
        ]

    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.siliconflow.cn')

        def on_response(response: Response):
            try:
                return default_on_response(response)
            except RawHTTPError as e:
                raise HTTPError.from_raw(e)

        self.on_response = on_response

    def chat(self, *user_prompt: Prompt):
        r = super().chat(*user_prompt, model=self.model, timeout=50, n=self.n)

        data = r['data']
        assert len(data) == self.n
        # siliconflow has no data structure in choices[0].message.annotations
        # See in https://docs.siliconflow.cn/cn/api-reference/chat-completions/chat-completions#response-choices-items-message
        assert len(r['annotations']) == 0
        #
        return data

    def which(self, query: str, documents: list[str], **kwargs) -> tuple[str, int]:
        json = {
            'model': self.model,
            'query': query,
            'documents': documents,
            **kwargs
        }
        response = self.request(f"{self.base_url}/rerank", "POST", json=json)
        most_relevant_index = max(response['results'], key=lambda x: x['relevance_score'])['index']

        return documents[most_relevant_index], most_relevant_index



from pathlib import Path
from typing import Literal

import requests
from davidkhala.utils.http_request import default_on_response
from requests import Response, HTTPError

from davidkhala.llm.api import API
from davidkhala.llm.model.file import FileAware

Purpose = Literal['voice_clone', 'prompt_audio', 't2a_async_input']
"""
See in https://platform.minimax.io/docs/api-reference/file-management-upload#body-purpose
    - voice_clone: For quick cloning of the original voice file (supports mp3, m4a, wav formats).
    - prompt_audio: Sample audio for voice cloning (supports mp3, m4a, wav formats).
    - t2a_async_input: Text file in the request body for asynchronous long-text-to-speech synthesis.(text-to-speech)
"""


class Minimax(API, FileAware):
    def __init__(self, api_key: str, base_url: str):
        super().__init__(api_key, base_url)

        def on_response(r: requests.Response) -> dict | None:
            r_dict = default_on_response(r)
            if r_dict is None:
                return None
            status = r_dict['base_resp']['status_code']
            msg = r_dict['base_resp']['status_msg']
            if status:
                derived = Response()
                derived.url = r.url
                derived.status_code = status
                derived.reason = msg
                derived.raise_for_status()
                raise HTTPError(f"{derived.status_code} Client Error: {derived.reason} for url: {derived.url}",
                                response=derived)
            else:
                assert status == 0
                assert msg == "success"
            return r_dict

        self.on_response = on_response

    def files_list(self):
        url = f"{self.base_url}/files/list"
        r = self.request(url, 'GET')
        return r['files']

    def file_get(self, file_id: int) -> dict | None:
        url = f"{self.base_url}/files/retrieve"
        try:
            r = self.request(url, 'GET', params={
                'file_id': file_id
            })
            return r['file']
        except HTTPError as err:
            if err.response.status_code == 2013 and err.response.reason == 'invalid params, file not found':
                return None
            raise err

    def files_delete(self, file_id: int, purpose: Purpose) -> int:
        url = f"{self.base_url}/files/delete"
        r = self.request(url, 'POST', json={
            "file_id": file_id,
            "purpose": purpose
        })

        return r['file_id']

    def nuke(self):
        r = []
        for _ in self.files_list():
            file_id = _['file_id']
            r.append(file_id)
            self.files_delete(_['file_id'], _['purpose'])

    def upload(self, file: Path, purpose: Purpose):
        """
        :param file:
        :param purpose:
        """
        url = f"{self.base_url}/files/upload"

        with open(file, 'rb') as f:
            r = self.request(url, 'POST', data={
                'purpose': purpose
            }, files={
                'file': (file.name, f)
            })
        return r['file']

class CN(Minimax):
    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.minimaxi.com')


class Global(Minimax):
    def __init__(self, api_key: str):
        super().__init__(api_key, 'https://api.minimax.io')



from davidkhala.llm.model import ModelAware


class Reranker(ModelAware):
    def which(self, query: str, documents: list[str], **kwargs) -> tuple[str, int]:
        """
        Returns:
            tuple:
                document (str): The most relevant document
                index (int): The index of most relevant document in documents

        """
        ...


from davidkhala.llm.anthropic import Client as AnthropicClient
from davidkhala.llm.model.chat import Prompt


class Client(AnthropicClient):
    def chat(self, *user_prompt: Prompt):
        """
        minimax supports text and tool calls, no image/document input
        """
        return super().chat(*user_prompt)


class Global(Client):
    def __init__(self, api_key: str):
        super().__init__(api_key=api_key, base_url='https://api.minimax.io/anthropic')


class CN(Client):
    def __init__(self, api_key: str):
        super().__init__(api_key=api_key, base_url='https://api.minimaxi.com/anthropic')



from time import sleep

import requests
from davidkhala.utils.http_request import default_on_response
from requests import Response

from davidkhala.llm.api import EmbeddingAPI, ChatAPI, GardenAPI
from davidkhala.llm.model.garden import TrialAvailable
from davidkhala.llm.model.chat import Prompt
from davidkhala.llm.model.openrouter import OpenRouterModel, Plugins


class OpenRouter(ChatAPI, EmbeddingAPI, GardenAPI, OpenRouterModel, TrialAvailable):
    def __init__(self, api_key: str, **kwargs):
        super().__init__(api_key=api_key, base_url='https://openrouter.ai/api')

        if 'leaderboard' in kwargs and type(kwargs['leaderboard']) is dict:
            # Site URL for rankings on openrouter.ai.
            self.options["headers"]["HTTP-Referer"] = kwargs['leaderboard']['url']
            # Site title for rankings on openrouter.ai.
            self.options["headers"]["X-Title"] = kwargs['leaderboard']['name']
        self.retry = True

        def on_response(response: requests.Response):
            """

            """
            r = default_on_response(response)
            # openrouter special error on response.ok
            err = r.get('error')
            if err:
                derived_response = Response()
                derived_response.status_code = err['code']
                derived_response.reason = err['message']
                derived_response.metadata = err.get("metadata")

                derived_response.raise_for_status()
            return r

        self.on_response = on_response

    def request(self, *args, **kwargs) -> dict:
        try:
            return super().request(*args, **kwargs)
        except requests.HTTPError as e:
            if e.response.status_code == 429 and self.retry:  # 429: You are being rate limited
                sleep(1)
                return self.request(*args, **kwargs)
            else:
                raise

    def chat(self, *user_prompt: Prompt, pdf_engine: Plugins.PDF_ENGINE = 'pdf-text'):
        options: dict = {
            'plugins': [
                Plugins.pdf(pdf_engine)
            ]
        }
        if self._models:
            options["models"] = self._models
        else:
            options["model"] = self.model

        r = super().chat(*user_prompt, **options)

        data, _a = r['data'], r['annotations']
        assert len(data) == OpenRouter.n
        assert len(_a) <= OpenRouter.n
        if _a:
            self.with_annotations(_a[0])
        if self._models:
            assert r['model'] in self._models
        return data[0]

    @property
    def models(self) -> list[str]:
        return [m['id'] for m in self.list_models()]

    @property
    def free_models(self) -> list[str]:
        return ['openrouter/free', *(_['id'] for _ in self.list_models() if _['id'].endswith(':free'))]
